require 'rails_helper'
require 'vcr_config'

module Hermes
  require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')
  require_relative '../../swagger_helper'

  ACCEPTABLE_PARAMS = %w[
    right
    type
    title
    group_id
    entity_id
    lessor_id
    user_id
  ]

  MUTATE_PARAMS = {
    right: ['reader', 'user'],
    type: ['entity', 'entity'],
    entity_id: ['entity1', 'entity2'],
    lessor_id: 'String',
    user_email: ['email1@test.com', 'email2@test.com'],
  }

  VALID_OPERATORS = %w[
    $in
    $nin
    $ne
    $gt
    $gte
    $lt
    $lte
    $regex
  ]

  describe Api::Dmo::UserRightsController, :type => :api do
    routes {Hermes::Engine.routes}
    include Devise::Test::ControllerHelpers
    include JwtAuthHelper
    
    check_mutable_params(
      model: Hermes::UserRight, 
      mutate_params: MUTATE_PARAMS, 
      factory: :user_right,
      other_factory:[
        {factory: :entity, id: 'entity1'},
        {factory: :entity, id: 'entity2'},
        {factory: :user, id: 'user1', email: 'email1@test.com'},
        {factory: :user, id: 'user2', email: 'email2@test.com'},
      ],
      min_params: {
        user_email: 'email1@test.com',
        right: 'reader',
        type: 'entity',
        entity_id: 'entity1',
        user_first_name: 'John', 
        user_last_name: 'Doe', 
        user_auth_roles: ['demandeur'] 
      },
      not_checked_keys: [:user_id, :user_first_name, :user_last_name, :user_auth_roles, :user_phone_number]
    )

    let!(:entity1) {create(:entity, siret: 'Super000000001', title: 'entity1')}
    let!(:entity2) {create(:entity, siret: 'Super000000002', title: 'entity2')}
    let!(:entity3) {create(:entity, siret: 'Super000000003', title: 'entity3')}
    let!(:entity4) {create(:entity, siret: 'Super000000003', title: 'entity4')}
    let!(:bailleur) { create(:user, auth_roles: ['bailleur'], first_name: 'Bailleur', email: 'bailleur@test.com') }
    let!(:user_right1) { create(:user_right, user_id: bailleur.id, entity: entity1.id, type: 'entity', right: 'administrator') }
    let!(:user_right2) { create(:user_right, user_id: bailleur.id, entity: entity2.id, type: 'entity', right: 'administrator') }
      
    let!(:user1) { create(:user, email: 'user1@test.com', auth_roles: ['demandeur'])}
    let!(:responsible) {create(:user, roles: ['Administrator'])}
    let!(:superadministrateur) {create(:user, roles: ['superadministrateur'])}


    describe 'POST /user_rights' do
      describe 'Create as bailleur' do
        it 'returns status code 201' do
          authenticated_header(request, bailleur)
          post 'create', {
            data: {
              attributes: { 
                right: 'administrator',
                type: 'entity',
                entity_id: entity1.id.to_s,
                user_email: user1.email,
              }
            }
          }
          expect(response).to have_http_status(201)
          expect(json['data']['attributes']['title']).to eq('entity1')
          expect(json['data']['attributes']['user_first_name']).to eq(user1.first_name)
          expect(Hermes::Entity.find_by(id: entity1.id).writable_by?(user1.reload)).to eq(true)
        end
        
      end

      describe 'Error when no rights' do
        before :each do
          authenticated_header(request, user1)
          post 'create', {
            data: {
              attributes: { 
                right: 'administrator',
                type: 'entity',
                entity_id: entity1.id.to_s,
                user_id: user1.id.to_s,
              }
            }
          }
        end
        it 'returns status code 401' do
          expect(response).to have_http_status(401)
        end
      end

      describe 'Create an new user' do
        before :each do
          authenticated_header(request, bailleur)
        end 

        it 'missing user_email' do
          post 'create', {
            data: {
              attributes: { 
                right: 'administrator',
                type: 'entity',
                entity_id: entity1.id.to_s,
              }
            }
          }
          expect(response).to have_http_status(422)
          expect(json['errors'].first['title']).to match(/missing_user_email/)
        end

        it 'missing missing_user_params' do
          post 'create', {
            data: {
              attributes: { 
                right: 'administrator',
                type: 'entity',
                entity_id: entity1.id.to_s,
                user_email: 'new_user@email.com',
              }
            }
          }
          expect(response).to have_http_status(422)
          expect(json['errors'].first['detail']).to eq("user params missing")
        end

        it 'error when creating user' do
          post 'create', {
            data: {
              attributes: { 
                right: 'administrator',
                type: 'entity',
                entity_id: entity1.id.to_s,
                user_email: '',
                user_first_name: 'John',
                user_last_name: 'Doe',
                user_auth_roles: ['bailleur'],
                user_gender: 'F'
              }
            }
          }
          expect(response).to have_http_status(422)
          expect(json['errors'].first['detail']).to match(/L'email ne peut pas être vide/)
        end

        it 'error already has right' do
          post 'create', {
            data: {
              attributes: { 
                right: 'reader',
                type: 'entity',
                entity_id: entity1.id.to_s,
                user_email: bailleur.email,
              }
            }
          }
          expect(response).to have_http_status(422)
          expect(json['errors'].first['detail']).to eq("This user already have a right on this entity administrator")
        end

        it 'works' do
          post 'create', {
            data: {
              attributes: { 
                right: 'administrator',
                type: 'entity',
                entity_id: entity1.id.to_s,
                user_email: 'new_user@email.com',
                user_first_name: 'John',
                user_last_name: 'Doe',
                user_auth_roles: ['bailleur'],
              }
            }
          }
          expect(response).to have_http_status(201)
          expect(json['data']['attributes']['title']).to eq('entity1')
          expect(json['data']['attributes']['right']).to eq('administrator')
          expect(json['data']['attributes']['type']).to eq('entity')
          expect(json['data']['attributes']['entity_id']).to eq(entity1.id.to_s)
          expect(json['data']['attributes']['user_email']).to eq('new_user@email.com')
          expect(json['data']['attributes']['user_first_name']).to eq('John')
          expect(json['data']['attributes']['user_last_name']).to eq('Doe')
          expect(json['data']['attributes']['user_auth_roles']).to eq(['bailleur'])
          expect(json['data']['attributes']['user_gender']).to eq('M')
          u = User.find_by(email: 'new_user@email.com')
          expect(u).not_to be_nil
          expect(u.gender).to eq('M')
        end
    
      end
    end

    describe 'PUT /user_rights/:id' do 
      describe 'as bailleur' do
        before :each do
          authenticated_header(request, bailleur)
          patch 'update', { id: user_right1.id,
            data: {
              attributes: { 
                right: 'reader',
              }
            }
          }
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
          expect(json['data']['attributes']['title']).to eq('entity1')
          expect(json['data']['attributes']['user_first_name']).to eq(bailleur.first_name)
          expect(Hermes::Entity.find_by(id: entity1.id).writable_by?(bailleur.reload)).to eq(false)
          expect(Hermes::Entity.find_by(id: entity1.id).readable_by?(bailleur.reload)).to eq(true)
        end
      end

      describe 'Cannot change user' do
        before :each do
          authenticated_header(request, bailleur)
          patch 'update', { id: user_right1.id,
            data: {
              attributes: { 
                user_id: user1,
              }
            }
          }
        end
        it 'returns status code 422' do
          expect(response).to have_http_status(422)
        end
      end

      describe 'Error when no rights' do
        before :each do
          authenticated_header(request, user1)
          patch 'update', { id: user_right1.id,
            data: {
              attributes: { 
                right: 'reader',
              }
            }
          }
        end
        it 'returns status code 401' do
          expect(response).to have_http_status(401)
        end
      end
    end 

    describe 'DELETE /user_rights/:id' do
      describe 'Delete as bailleur' do
        before :each do
          authenticated_header(request, bailleur)
          delete 'destroy', {
            id: user_right1.id.to_s
          }
        end
        it 'returns status code 204' do
          expect(response).to have_http_status(204)
          expect(Hermes::Entity.find_by(id: entity1.id).writable_by?(bailleur.reload)).to eq(false)
        end
      end

      describe 'Error when no rights' do
        before :each do
          authenticated_header(request, user1)
          delete 'destroy', {
            id: user_right1.id.to_s
          }
        end
        it 'returns status code 401' do
          expect(response).to have_http_status(401)
        end
      end
    end

    describe 'DELETE /user_rights/all_from_user' do

      it 'Delete as bailleur returns status code 204' do
        authenticated_header(request, bailleur)
        delete 'delete_all_from_user', { user_id: bailleur.id.to_s }
        expect(response).to have_http_status(204)
        expect(Hermes::Entity.writable_by(bailleur.reload).count).to eq(0)
      end

      it 'Error when missing user_id' do
        authenticated_header(request, bailleur)
        delete 'delete_all_from_user', { id: user_right1.id.to_s }
        expect(response).to have_http_status(422)
        expect(Hermes::Entity.writable_by(bailleur.reload).count).to eq(2)
      end

      it 'Error when no right' do
        authenticated_header(request, user1)
        delete 'delete_all_from_user', { user_id: bailleur.id.to_s }
        expect(response).to have_http_status(401)
        expect(Hermes::Entity.writable_by(bailleur.reload).count).to eq(2)
      end

    end

    describe 'GET /user_rights/:id' do
      it 'when access' do
        authenticated_header(request, bailleur)
        get 'show', id: user_right1.id.to_s
        datas = json['data']
        expect(response).to have_http_status(200)
        expect(datas['attributes']['title']).to eq entity1.title
      end

      it 'when superadministrateur' do
        authenticated_header(request, superadministrateur)
        get 'show', id: user_right1.id.to_s
        expect(response).to have_http_status(200)
      end

      it 'when no access' do
        authenticated_header(request, user1)
        get 'show', id: user_right1.id.to_s
        expect(response).to have_http_status(404)
      end
    end

    describe 'GET /user_rights' do
      describe 'Index as bailleur' do
        before :each do
          authenticated_header(request, bailleur)
          get 'index'
        end
        it 'returns status code 200' do
          datas = json['data']
          expect(response).to have_http_status(200)
          expect(datas.length).to eq 2
          expect(datas[0]['attributes']['title']).to eq entity1.title
        end
      end

      describe 'Index as bailleur filter by entity_id' do
        it 'returns status code 200' do
          authenticated_header(request, bailleur)
          get 'index', {entity_id: entity1.id.to_s}
          datas = json['data']
          expect(response).to have_http_status(200)
          expect(datas.length).to eq 1
        end
      end

      it 'work when superadministrateur' do 
        authenticated_header(request, superadministrateur)
        get 'index'
        expect(response).to have_http_status(200)
        expect(json['data'].length).to eq(2)
      end 
      
      it 'work when acces to entity' do 
        authenticated_header(request, bailleur)
        get 'index'
        expect(response).to have_http_status(200)
        expect(json['data'].length).to eq(2)
      end 

      it 'work when no acces to entity' do 
        authenticated_header(request, user1)
        get 'index'
        expect(response).to have_http_status(200)
        expect(json['data'].length).to eq(0)
      end 
    end

    describe 'POST /user_rights/search' do 
      it 'works' do 
        authenticated_header(request, bailleur)
        post 'search', {entity_id: entity1.id.to_s}
        datas = json['data']
        expect(response).to have_http_status(200)
        expect(datas.length).to eq 1
      end 
    end 

    describe 'DELETE /user_rights/bulk' do 
      before do 
        authenticated_header(request, bailleur)
      end 
      it 'works' do 
        delete 'delete_bulk', {user_right_ids: [user_right1.id.to_s]}
        expect(response).to have_http_status(204)
        expect(Hermes::UserRight.all.count).to eq(1)
      end 

      it 'when not found' do 
        delete 'delete_bulk', {user_right_ids: [user_right1.id.to_s, "dsqdqs"]}
        expect(response).to have_http_status(404)
      end 

      it 'when not right' do 
        ur = Hermes::UserRight.create(
          user_id: user1.id,
          entity_id: entity4.id,
          type: 'entity',
          right: 'administrator'
        )
        delete 'delete_bulk', {user_right_ids: [user_right1.id.to_s, ur.id.to_s]}
        expect(response).to have_http_status(401)
      end 

    end 

    describe 'POST /user_rights/bulk' do 
      let!(:params) do 
        {
          user_rights: [
            {user_email: user1.email, entity_id: entity1.id.to_s, type: 'entity', right: 'administrator'},
            {user_email: user1.email, entity_id: entity2.id.to_s, type: 'entity', right: 'administrator'},
          ]
        }
      end 

      let!(:params_new_user) do 
        {
          user_rights: [
            {
              user_email: "test@email.com", 
              entity_id: entity1.id.to_s, 
              type: 'entity', 
              right: 'administrator',
              user_first_name: 'John',
              user_last_name: 'Doe',
              user_auth_roles: ['entreprise'],
              user_gender: 'F'
            },
          ]
        }
      end 

      before do 
        authenticated_header(request, bailleur)
      end 

      it 'works' do 
        expect(Hermes::UserRight.where(user: user1).count).to eq(0)
        post 'post_bulk', params
        datas = json['data']
        expect(response).to have_http_status(200)
        expect(datas.length).to eq 2
        expect(user1.reload.group_ids.length).to eq(2)
        expect(Hermes::UserRight.where(user: user1).count).to eq(2)
      end
      
      it 'when missing params' do 
        params[:user_rights][0].delete(:user_email)
        post 'post_bulk', params
        expect(response).to have_http_status(422)
      end 

      it 'when no right' do 
        params[:user_rights][0][:entity_id] = entity3.id.to_s
        post 'post_bulk', params
        expect(response).to have_http_status(401)
      end 

      it 'works when creating new user' do 
        post 'post_bulk', params_new_user
        datas = json['data']
        expect(response).to have_http_status(200)
        expect(datas.length).to eq 1
      end 

      it 'works when creating new user with error' do 
        post 'post_bulk', params[:user_rights][0].delete(:user_first_name)
        expect(response).to have_http_status(422)
        expect(json['errors'].first['title']).to match(/missing_params/)
      end 

    end 

    describe 'GET /user_rights/is_administrator_for_lessor' do 
      it 'works for admin' do 
        authenticated_header(request, bailleur)
        get 'is_administrator_for_entity'
        expect(response).to have_http_status(200)
        expect(json['is_administrator']).to eq(true)
      end 

      it 'doesnt work for lessor' do 
        authenticated_header(request, bailleur)
        get 'is_administrator_for_lessor'
        expect(response).to have_http_status(200)
        expect(json['is_administrator']).to eq(false)
      end 

      it 'works with user params' do 
        authenticated_header(request, responsible)
        get 'is_administrator_for_entity', user_id: bailleur.id
        expect(response).to have_http_status(200)
        expect(json['is_administrator']).to eq(true)
      end 

      it 'doesnt work with not found user' do 
        authenticated_header(request, responsible)
        get 'is_administrator_for_entity', user_id: "unknown"
        expect(response).to have_http_status(404)
      end 
    end 
  end
end
